#!/usr/bin/env python

"""
Search Python abstract syntax trees for nodes of a particular type having a
particular textual value.

Copyright (C) 2008 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import compiler
import os
import linecache
import types

__version__ = "0.1.1"

# Excluded AST nodes and their names.

excluded_term_types = ["Module", "Stmt"]
excluded_term_cls = tuple([getattr(compiler.ast, name) for name in excluded_term_types])

# Search functions.

def search_recursive(directory, term_type, op=None):

    """
    Search files within the filesystem below 'directory' for terms having the
    given 'term_type', using 'op' (if specified) to match a search term.
    """

    results = []
    for path, directories, filenames in os.walk(directory):
        for filename in filenames:
            if os.path.splitext(filename)[-1] == os.path.extsep + "py":
                results += search_file(os.path.join(path, filename), term_type, op)
    return results

def search_file(filename, term_type, op=None):

    """
    Search the file with the given 'filename' for terms having the given
    'term_type', using 'op' (if specified) to match a search term. If
    'term_type' is given as "*", attempt to match any term type.
    """

    try:
        node = compiler.parseFile(filename)
    except SyntaxError:
        return []

    if term_type != "*":
        cls = getattr(compiler.ast, term_type)
    else:
        cls = None

    return search_tree(node, cls, op, filename)

def search_tree(node, cls, op=None, filename=None):

    """
    Search the tree rooted at the given 'node' for nodes of the given class
    'cls', using 'op' (if specified) to match a search term. If 'cls' is None,
    all node types will be considered for matches.

    Return a list of results of the form (node, value, filename).
    """

    results = []

    # Ignore excluded nodes.

    if isinstance(node, excluded_term_cls):
        pass

    # Test permitted nodes.

    elif cls is None or isinstance(node, cls):
        if op is None:
            results.append((node, None, filename))
        else:
            found_str = 0

            for child in node.getChildren():

                # Test literals.

                if isinstance(child, (int, float, long, bool)):
                    if op(str(child)):
                        results.append((node, child, filename))

                # Only check a single string child value since subsequent
                # values are typically docstrings.

                elif not found_str and isinstance(child, str):
                    found_str = 1
                    if op(child):
                        results.append((node, child, filename))

                # Argument lists, globals and imports.

                elif isinstance(child, list):
                    results += search_list(child, node, op, filename)

    # Search within nodes, even if matches have already been found.

    for child in node.getChildNodes():
        results += search_tree(child, cls, op, filename)

    return results

def search_list(values, node, op=None, filename=None):

    """
    Search the given 'values' from the given 'node', using 'op' (if specified)
    to match a search term.

    Return a list of results of the form (node, value, filename).
    """

    results = []

    for value in values:

        # Test strings.

        if isinstance(value, str) and op(str(value)):
            results.append((node, value, filename))

        # Test import tuples.

        elif isinstance(value, tuple):
            for subvalue in value:
                if isinstance(value, str) and op(str(subvalue)):
                    results.append((node, subvalue, filename))

    return results

def expand_results(results):

    """
    Expand the given 'results', making a list containing tuples of the form
    (node, filename, line number, line, value).
    """

    expanded = []

    for node, value, filename in results:
        lineno = node.lineno

        if filename is not None and lineno is not None:
            line = linecache.getline(filename, lineno).rstrip()
        else:
            line = None

        expanded.append((node, filename, lineno, line, value))

    return expanded

def get_term_types():

    "Return the term types supported by the module."

    term_types = []

    for name in dir(compiler.ast):
        if name in excluded_term_types:
            continue

        obj = getattr(compiler.ast, name)

        if isinstance(obj, types.ClassType) and \
            issubclass(obj, compiler.ast.Node) and \
            name[0].isupper():

            term_types.append(name)

    return term_types

# Command syntax.

syntax_description = """
    [ -n | --line-number ]
    [ -p | --print-token ]
    [ ( -t TERM_TYPE ) | ( --type=TERM_TYPE ) ]
    [ ( -e PATTERN ) | ( --regexp=PATTERN ) ]
    [ -r | -R | --recursive ] ( FILENAME ... )
    """

# Main program.

def run_command():

    "The functionality of the main program."

    import sys
    import cmdsyntax
    import re
    import textwrap

    # Match command arguments.

    syntax = cmdsyntax.Syntax(syntax_description)
    syntax_matches = syntax.get_args(sys.argv[1:])
    show_syntax = 0

    try:
        args = syntax_matches[0]
    except IndexError:
        show_syntax = 1

    if show_syntax:
        print "Syntax:"
        print syntax_description
        print "Term types:"
        print "\n".join(textwrap.wrap(", ".join(get_term_types())))
        sys.exit(1)

    # Get the search details.

    term_type = args.get("TERM_TYPE", "*")
    term = args.get("PATTERN")
    recursive = args.has_key("r") or args.has_key("R") or args.has_key("recursive")

    if term is None:
        op = None
    else:
        op = re.compile(term).search

    # Perform the search in files and directory hierarchies.

    results = []

    for filename in args["FILENAME"]:
        if os.path.isfile(filename):
            results += search_file(filename, term_type, op)
        elif recursive and os.path.isdir(filename):
            results += search_recursive(filename, term_type, op)

    # Present the results.

    for node, filename, lineno, line, value in expand_results(results):
        format = "%s:"
        output = [filename]

        # Handle line numbers and missing details.

        if args.has_key("n") or args.has_key("line-number"):
            if lineno is not None:
                format += "%d:"
                output.append(lineno)

        # Show matching tokens, if requested.

        if args.has_key("p"):
            if value is not None:
                format += "%r:"
                output.append(value)
            else:
                format += "%s:"
                output.append("<%s>" % (term_type or "*"))

        # Show lines, if defined.

        if line is not None:
            format += " %s"
            output.append(line)

        print format % tuple(output)

if __name__ == "__main__":
    run_command()

# vim: tabstop=4 expandtab shiftwidth=4
